import subprocess
import sys

RESULTS = {
    "WORST": "_R___",
    "HOUSE": "RR___",
    "WORSE": "__RW_",
    "ZONES": "RR___",
    "HORSE": "B_WR_",
    "EBONY": "RRWRR",
    "LOGAN": "__R__",
    "FALSE": "__W__",
    "ARGON": "_RRR_",
    "GORSE": "____R",
    "HORDE": "_R__W",
    "MORSE": "WRRR_",
    "NORSE": "___WR",
    "EVILS": "__R__",
    "WINCE": "___R_",
    "WHOSE": "_RR__",
    "WASTE": "_R__R",
    "SHORE": "_____",
    "HYPER": "____R",
    "HYENA": "____R",

    "HIDER": "_R_R_",
    "HIKES": "_R_R_",
    "LURED": "__RWR",
    "FLESH": "_____",
    "HUGER": "_R_R_",
    "HEFTS": "__R__",
    "SOLID": "___RR",
    "HYDRA": "__R__",
    "SHRUB": "___RR",
    "LEASH": "_RBR_",
    "PERTH": "_____",
    "HERTZ": "_____",
    "DEIST": "___R_",
    "QUEST": "___RR",
    "CHEST": "_R___",
    "HELIX": "___R_",
    "HERZL": "_____",
    "PRATE": "B___R",
    "FERAL": "_R___",
    "CLARE": "_____",
    "PERIL": "R__W_",
    "HEART": "BBBBB",
}

WORDS = set([w.strip().upper() for w in open("/usr/share/dict/american-english") if len(w.strip()) == 5 and all(l in "abcdefghijklmnopqrstuvwxyz" for l in w.lower().strip())])

header = """
(declare-datatypes () ((letter A B C D E F G H I J K L M N O P Q R S T U V W X Y Z)))
(declare-const l0 letter)
(declare-const l1 letter)
(declare-const l2 letter)
(declare-const l3 letter)
(declare-const l4 letter)
(assert (distinct l0 l1 l2 l3 l4))

(define-fun countb ((w0 letter) (w1 letter) (w2 letter) (w3 letter) (w4 letter)) Int
  (+ (ite (= l0 w0) 1 0)
     (ite (= l1 w1) 1 0)
     (ite (= l2 w2) 1 0)
     (ite (= l3 w3) 1 0)
     (ite (= l4 w4) 1 0)))

(define-fun countr ((w0 letter) (w1 letter) (w2 letter) (w3 letter) (w4 letter)) Int
  (+ (ite (distinct w0 l0 l1 l2 l3 l4) 1 0)
     (ite (distinct w1 l0 l1 l2 l3 l4) 1 0)
     (ite (distinct w2 l0 l1 l2 l3 l4) 1 0)
     (ite (distinct w3 l0 l1 l2 l3 l4) 1 0)
     (ite (distinct w4 l0 l1 l2 l3 l4) 1 0)))

(define-fun countw ((w0 letter) (w1 letter) (w2 letter) (w3 letter) (w4 letter)) Int
  (- 5 (countb w0 w1 w2 w3 w4) (countr w0 w1 w2 w3 w4)))
"""

body = []

def word_constraints(word, result):
    word = word.upper()
    result = result.upper()
    numb = result.count("B")
    numw = result.count("W")
    numr = result.count("R")
    num_ = result.count("_")
    assert numb + numw + numr + num_ == len(result)

    body = []
    body.append("(assert (<= {} (countb {}) {}))".format(numb, " ".join(word), numb + num_))
    body.append("(assert (<= {} (countw {}) {}))".format(numw, " ".join(word), numw + num_))
    body.append("(assert (<= {} (countr {}) {}))".format(numr, " ".join(word), numr + num_))
    return "\n".join(body)

def print_wordlist(fin, fout):
    fin.write("(assert (or " + 
              " ".join(
                  "(and " + " ".join("(= l{} {})".format(i, l) for i, l in enumerate(word)) + ")"
                  for word in WORDS) + "))\n") 
    fin.flush()

def talk(fin, fout):
    fin.write(header)
    fin.flush()

    print_wordlist(fin, fout)
    
    for word, result in RESULTS.items():
        fin.write(word_constraints(word, result) + "\n")
        fin.flush()

    while True:
        word = get_model(fin, fout)
        if word:
            result = input(word + " > ").strip()
            RESULTS[word] = result
            fin.write("(assert (not (and " + " ".join("(= l{} {})".format(i, l) for i, l in enumerate(word)) + " )))\n")
            fin.write(word_constraints(word, result) + "\n")
            fin.flush()

def get_model(fin, fout):
    fin.write("(check-sat)\n")
    fin.flush()

    is_sat = fout.readline().strip()
    assert is_sat == "sat", is_sat
    fin.write("(get-model)\n")
    fin.flush()

    out = {}
    fout.readline()
    for i in range(5):
        i = fout.readline()[15]
        l = fout.readline()[4]
        out[i] = l
    fout.readline()
    word = out["0"] + out["1"] + out["2"] + out["3"] + out["4"]

    if word not in WORDS:
        fin.write("(assert (not (and " + " ".join("(= l{} {})".format(i, l) for i, l in enumerate(word)) + ")))\n")
        fin.flush()
        print("# Not", word)
        return False
    else:
        return word

if __name__ == "__main__":
    z3 = subprocess.Popen("z3 -smt2 -in", shell=True, encoding="utf8", stdin=subprocess.PIPE, stdout=subprocess.PIPE)
    talk(z3.stdin, z3.stdout)
